import numpy as np

def get_coo_arrays(n_verts, faces, lambda_):

    # Neighbor indices
    ii = faces[:, [1, 2, 0]].flatten()
    jj = faces[:, [2, 0, 1]].flatten()
    adj = np.unique(np.stack([np.concatenate([ii, jj]), np.concatenate([jj, ii])], axis=0), axis=1)
    adj_values = np.ones(adj.shape[1], dtype=np.float64) * lambda_

    # Diagonal indices, duplicated as many times as the connectivity of each index
    diag_idx = np.stack((adj[0], adj[0]), axis=0)

    diag = np.stack((np.arange(n_verts), np.arange(n_verts)), axis=0)

    # Build the sparse matrix
    idx = np.concatenate((adj, diag_idx, diag), axis=1)
    values = np.concatenate((-adj_values, adj_values, np.ones(n_verts)))

    return values, idx

# https://sinestesia.co/blog/tutorials/python-icospheres/
def get_icosphere(level=0):
    # Generate the base, that we then subdivide
    n_verts = 12

    faces = [
            [0, 11, 5],
            [0, 5, 1],
            [0, 1, 7],
            [0, 7, 10],
            [0, 10, 11],

            [1, 5, 9],
            [5, 11, 4],
            [11, 10, 2],
            [10, 7, 6],
            [7, 1, 8],

            [3, 9, 4],
            [3, 4, 2],
            [3, 2, 6],
            [3, 6, 8],
            [3, 8, 9],

            [4, 9, 5],
            [2, 4, 11],
            [6, 2, 10],
            [8, 6, 7],
            [9, 8, 1]]

    split_cache = {}

    def split(i, j):
        key = f"{min(i,j)}-{max(i,j)}"

        if key in split_cache:
            return split_cache[key]

        split_cache[key] = n_verts

        return n_verts

    for i in range(level):
        faces_subdiv = []

        for tri in faces:
            v1 = split(tri[0], tri[1])
            n_verts = max(n_verts, v1+1)
            v2 = split(tri[1], tri[2])
            n_verts = max(n_verts, v2+1)
            v3 = split(tri[2], tri[0])
            n_verts = max(n_verts, v3+1)

            faces_subdiv.append([tri[0], v1, v3])
            faces_subdiv.append([tri[1], v2, v1])
            faces_subdiv.append([tri[2], v3, v2])
            faces_subdiv.append([v1, v2, v3])

        faces = faces_subdiv

    return n_verts, np.array(faces)

def get_cube():
    faces = np.array([[0, 1, 3],
                     [0, 3, 2],
                     [2, 3, 7],
                     [2, 7, 6],
                     [4, 5, 7],
                     [4, 7, 6],
                     [0, 1, 5],
                     [0, 5, 4],
                     [1, 5, 7],
                     [1, 7, 3],
                     [0, 4, 6],
                     [0, 6, 2]])
    return 8, faces



import pytest
from cholespy import CholeskySolverD, CholeskySolverF, MatrixType
import numpy as np
import scipy.sparse as sp
from scipy.sparse.linalg import spsolve
from utils import get_coo_arrays, get_icosphere, get_cube

RHS = [1, 4, 16, 64, 128, 256]

@pytest.mark.parametrize("n_verts, faces", [get_cube(), get_icosphere(3), get_icosphere(5)])
@pytest.mark.parametrize("variant", ["float", "double"])
@pytest.mark.parametrize("device", ["cpu", "cuda"])
def test_solver(device, variant, n_verts, faces):
    try:
        import torch
    except ModuleNotFoundError:
        pytest.skip("PyTorch was not found!")

    if device=='cuda' and torch.cuda.device_count() == 0:
        pytest.skip("No GPU found.")

    np.random.seed(45)
    CholeskySolver = CholeskySolverF if variant == "float" else CholeskySolverD
    dtype = np.float32 if variant == "float" else np.float64

    lambda_ = 2.0
    values, idx = get_coo_arrays(n_verts, faces, lambda_)

    L_csc = sp.csc_matrix((values, idx))
    b = np.random.random(size=(n_verts, max(RHS))).astype(dtype)
    x= spsolve(L_csc, b)
    solver = CholeskySolver(n_verts, torch.tensor(idx[0], device=device), torch.tensor(idx[1], device=device), torch.tensor(values, device=device), MatrixType.COO)

    # Test with different RHS
    for n_rhs in RHS:
        b_crop = b[:, :n_rhs]
        b_torch = torch.tensor(b_crop, device=device)
        x_torch = torch.zeros_like(b_torch)
        try:
            solver.solve(b_torch, x_torch)
            assert(np.allclose(x_torch.cpu().numpy(), x[:, :n_rhs]))
        except ValueError as e:
            assert n_rhs > 128
            assert e.__str__() == "The number of RHS should be less than 128."

@pytest.mark.parametrize("device", ["cpu", "cuda"])
def test_matrices(device):
    try:
        import torch
    except ModuleNotFoundError:
        pytest.skip("PyTorch was not found!")

    if device=='cuda' and torch.cuda.device_count() == 0:
        pytest.skip("No GPU found.")

    np.random.seed(45)
    lambda_ = 2.0
    n_verts, faces = get_cube()
    values, idx = get_coo_arrays(n_verts, faces, lambda_)

    L_csc = sp.csc_matrix((values, idx))
    L_csr = sp.csr_matrix((values, idx))

    b = np.random.random(size=(n_verts, 32)).astype(np.float32)
    b_torch = torch.tensor(b, device=device)
    x_torch = torch.zeros_like(b_torch)

    x_ref = spsolve(L_csc, b)

    # Test with COO input
    solver = CholeskySolverF(n_verts, torch.tensor(idx[0], device=device), torch.tensor(idx[1], device=device), torch.tensor(values, device=device), MatrixType.COO)
    solver.solve(b_torch, x_torch)

    assert(np.allclose(x_torch.cpu().numpy(), x_ref))

    # Test with CSR input
    solver = CholeskySolverF(n_verts, torch.tensor(L_csr.indptr, device=device), torch.tensor(L_csr.indices, device=device), torch.tensor(L_csr.data, device=device), MatrixType.CSR)
    solver.solve(b_torch, x_torch)

    assert(np.allclose(x_torch.cpu().numpy(), x_ref))

    # Test with CSC input
    solver = CholeskySolverF(n_verts, torch.tensor(L_csc.indptr, device=device), torch.tensor(L_csc.indices, device=device), torch.tensor(L_csc.data, device=device), MatrixType.CSC)
    solver.solve(b_torch, x_torch)

    assert(np.allclose(x_torch.cpu().numpy(), x_ref))

@pytest.mark.parametrize("framework", ["numpy", "torch", "tensorflow", "jax", "cupy", "drjit"])
def test_frameworks(framework):
    import importlib
    # Disable tests if the module is not installed
    try:
        importlib.import_module(framework)
    except ModuleNotFoundError:
        pytest.skip(f"Module {framework} is not installed.")

    np.random.seed(45)
    lambda_ = 2.0

    n_verts, faces = get_cube()
    values, idx = get_coo_arrays(n_verts, faces, lambda_)

    L_csc = sp.csc_matrix((values, idx))
    b = np.random.random(size=(n_verts, 32)).astype(np.float32)
    x_ref = spsolve(L_csc, b)

    # Test with Numpy
    if framework == "numpy":
        solver = CholeskySolverF(n_verts, idx[0], idx[1], values, MatrixType.COO)

        x = np.zeros_like(b)
        solver.solve(b, x)
        assert(np.allclose(x, x_ref))

    elif framework == "torch":
        import torch
        # Test with PyTorch - CPU
        solver = CholeskySolverF(n_verts, torch.tensor(idx[0]), torch.tensor(idx[1]), torch.tensor(values), MatrixType.COO)

        b_torch = torch.tensor(b)
        x_torch = torch.zeros_like(b_torch)
        solver.solve(b_torch, x_torch)
        assert(np.allclose(x_torch.numpy(), x_ref))

        # Test with PyTorch - CUDA
        if torch.cuda.device_count() > 0:
            solver = CholeskySolverF(n_verts, torch.tensor(idx[0], device='cuda'), torch.tensor(idx[1], device='cuda'), torch.tensor(values, device='cuda'), MatrixType.COO)

            b_torch = torch.tensor(b, device='cuda')
            x_torch = torch.zeros_like(b_torch)
            solver.solve(b_torch, x_torch)
            assert(np.allclose(x_torch.cpu().numpy(), x_ref))

    elif framework == "tensorflow":
        import tensorflow as tf
        gpus = tf.config.experimental.list_physical_devices('GPU')
        # Prevent TF from allocating all GPU mrmory for itself
        for gpu in gpus:
            tf.config.experimental.set_memory_growth(gpu, True)
        # Test with TensorFlow - CPU
        with tf.device('/device:cpu:0'):
            solver = CholeskySolverF(n_verts, tf.convert_to_tensor(idx[0]), tf.convert_to_tensor(idx[1]), tf.convert_to_tensor(values), MatrixType.COO)

            b_tf = tf.convert_to_tensor(b)
            x_tf = tf.zeros_like(b_tf)
            solver.solve(b_tf, x_tf)
            assert(np.allclose(x_tf.numpy(), x_ref))

        if len(gpus) > 0:
            # Test with TensorFlow - CUDA
            with tf.device('/device:gpu:0'):
                solver = CholeskySolverF(n_verts, tf.convert_to_tensor(idx[0]), tf.convert_to_tensor(idx[1]), tf.convert_to_tensor(values), MatrixType.COO)

                b_tf = tf.convert_to_tensor(b)
                x_tf = tf.zeros_like(b_tf)
                solver.solve(b_tf, x_tf)
                assert(np.allclose(x_tf.numpy(), x_ref))

    elif framework == "jax":
        import os
        # Prevent JAX from allocating all GPU mrmory for itself
        os.environ['XLA_PYTHON_CLIENT_PREALLOCATE'] = "false"
        import jax
        with jax.experimental.enable_x64():
            # Test with JAX
            solver = CholeskySolverF(n_verts, jax.numpy.array(idx[0]), jax.numpy.array(idx[1]), jax.numpy.array(values, dtype=np.float64), MatrixType.COO)

            b_jax= jax.numpy.array(b)
            x_jax = jax.numpy.zeros_like(b_jax)
            solver.solve(b_jax, x_jax)
            assert(np.allclose(x_jax, x_ref))

    elif framework == "cupy":
        import cupy as cp
        # Test with CuPy
        solver = CholeskySolverF(n_verts, cp.array(idx[0], dtype=cp.int32), cp.array(idx[1], dtype=cp.int32), cp.array(values, dtype=cp.float64), MatrixType.COO)

        b_cp = cp.array(b)
        x_cp = cp.zeros_like(b_cp)
        solver.solve(b_cp, x_cp)
        assert(np.allclose(cp.asnumpy(x_cp), x_ref))

    elif framework == "drjit":
        import drjit
        # Test with DrJIT - CUDA
        solver = CholeskySolverF(n_verts, drjit.cuda.TensorXi(idx[0]), drjit.cuda.TensorXi(idx[1]), drjit.cuda.TensorXf64(values), MatrixType.COO)

        b_drjit = drjit.cuda.TensorXf(b)
        x_drjit = drjit.zeros(drjit.cuda.TensorXf, b.shape)
        solver.solve(b_drjit, x_drjit)
        assert(np.allclose(x_drjit.numpy(), x_ref))

        # Test with DrJIT - CPU
        solver = CholeskySolverF(n_verts, drjit.llvm.TensorXi(idx[0]), drjit.llvm.TensorXi(idx[1]), drjit.llvm.TensorXf64(values), MatrixType.COO)

        b_drjit = drjit.llvm.TensorXf(b)
        x_drjit = drjit.zeros(drjit.llvm.TensorXf, b.shape)
        solver.solve(b_drjit, x_drjit)
        assert(np.allclose(x_drjit.numpy(), x_ref))



#!/usr/bin/python
# Generate "wrapper sources" for compiling the same file several times with CMake for UMFPACK
# (Jose Luis Blanco)

import os
import sys
import re
import string

#------   MAIN   -------
def main():
	f = open("makefile2wrappers.txt","r");
	lins = f.readlines();
	f.close();
	
	for l in lins:
		l = l.strip();
		if len(l)==0 or l.startswith('#'):
			continue;

		print('Line: '+l);
		# $(C) -DDINT -c ../Source/umf_analyze.c -o umf_i_analyze.o
		defs=re.match(".*\)(.*)-c",l).group(1).strip();
		# If there's no "-o" flag, just compile the file as is:
		if re.search('.*-o.*',l)!=None:
			src=re.match(".*-c(.*)-o",l).group(1).strip();
			out=re.match(".*-o(.*)",l).group(1).strip();
			f='SourceWrappers/'+out+".c";
			print(' => Creating '+f+'\n');
			o = open(f,"w");
			DEFs = defs.strip().split("-D");
			DEFs = [x for x in DEFs if x]; # Remove empty 
			# If we have "CS_COMPLEX", wrap the entire thing in a #ifndef NCOMPLEX ... #endif
			has_complex=False
			if 'CS_COMPLEX' in DEFs:
				has_complex=True
			# TODO: Other COMPLEX flags?
			
			if has_complex:
				o.write('#ifndef NCOMPLEX\n');
			# Pre-definitions:
			for d in DEFs:
				o.write('#define '+d+'\n');
			# The source code itself:
			o.write('#include <'+src+'>'+'\n');
			if has_complex:
				o.write('#endif // NCOMPLEX\n');
			o.close();	
		else:
			src=re.match(".*-c(.*)",l).group(1).strip();
			f = "SourceWrappers/"+os.path.basename(src);
			print(' => Creating '+f+'\n');
			o = open(f,"w");
			o.write('#include <'+src+'>'+'\n');
			o.close();
				
	return 0
#-----------


def replaceInFile(fileIn, fileOut, textToSearch, newText) :
    o = open(fileOut,"w")
    data = open(fileIn).read()
    for idx in range(len(textToSearch)):
        data = re.sub(textToSearch[idx], newText[idx], data)
    o.write(data)
    o.close()
#-----------
	
	
if __name__ == "__main__":
    sys.exit(main())




import sys

try:
    from skbuild import setup
    import nanobind
except ImportError:
    print("The preferred way to invoke 'setup.py' is via pip, as in 'pip "
          "install .'. If you wish to run the setup script directly, you must "
          "first install the build dependencies listed in pyproject.toml!",
          file=sys.stderr)
    raise

from setuptools import find_packages

# read the contents of your README file (https://packaging.python.org/en/latest/guides/making-a-pypi-friendly-readme/)
from pathlib import Path
this_directory = Path(__file__).parent
readme = (this_directory / "README.md").read_text()

setup(
    name="cholespy",
    version="0.1.5",
    description="A self-contained sparse Cholesky solver, compatible with CPU and GPU tensor frameworks.",
    author="Baptiste Nicolet",
    license="BSD",
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    cmake_install_dir="src/cholespy",
    include_package_data=True,
    python_requires=">=3.8",
    long_description=readme,
    long_description_content_type="text/markdown"
)
